from __future__ import annotations

import numpy as np
import multiprocessing
import scipy.signal
from functools import partial
from typing import Callable

try:
    import simpleaudio as sa
except ModuleNotFoundError:
    raise ModuleNotFoundError("Please install simpleaudio to use this module.\n"
                              "pip install simpleaudio\n\n"
                              "For Linux see simpleaudio dependecies at:\n"
                              "https://simpleaudio.readthedocs.io"
                              "/en/latest/installation.html#linux-dependencies")


MAX_FREQ = 2000

SignalFunction = Callable[[np.ndarray], np.ndarray]


def sine(x: np.ndarray, freq: float, phase: float) -> np.ndarray:
    return np.sin(x * 2 * np.pi * freq + phase)


def triangular(x: np.ndarray, freq: float, phase: float) -> np.ndarray:
    return scipy.signal.sawtooth(x * 2 * np.pi * freq + phase, width=0.5)


def square(x: np.ndarray, freq: float, phase: float) -> np.ndarray:
    return scipy.signal.square(x * 2 * np.pi * freq + phase, duty=0.5)


def make_signal(f: SignalFunction, duration: float, fs: int = 44100):
    step = 1/fs
    n = duration // step
    t = np.arange(n) * step
    return f(t)


class WaveGeneratorChannel:
    def __init__(self, fs: int = 44100, duration: int = 60):
        self._duration = duration
        self._fs = fs
        self._freq = 440
        self._waveform = "sine"
        self._phase = 0
        self._amplitude = 1.0
        self._enabled_output = False
        self._enabled_controls = True
        self._waveforms = {
            "sine": sine,
            "triangular": triangular,
            "square": square
        }

    @property
    def enabled(self) -> bool:
        return self._enabled_output

    @enabled.setter
    def enabled(self, value: bool):
        self._enabled_output = value

    @property
    def frequency(self) -> int:
        return self._freq

    @frequency.setter
    def frequency(self, value: int):
        if self._enabled_controls is False:
            raise RuntimeError("Controls of WaveGenerator are blocked during loop")
        if 1 <= value <= MAX_FREQ:
            self._freq = int(value)
        else:
            raise ValueError(f"MIN: 1, MAX: {MAX_FREQ}")

    @property
    def phase(self) -> float:
        return self._phase

    @phase.setter
    def phase(self, value: float):
        if self._enabled_controls is False:
            raise RuntimeError("Controls of WaveGenerator are blocked during loop")

        self._phase = (value - np.pi) % (2 * np.pi) - np.pi

    @property
    def amplitude(self) -> float:
        return self._amplitude

    @amplitude.setter
    def amplitude(self, value: float):
        if self._enabled_controls is False:
            raise RuntimeError("Controls of WaveGenerator are blocked during loop")
        if 0.0 <= value <= 1.0:
            self._amplitude = float(value)
        else:
            raise ValueError(f"MIN: 0.0, MAX: 1.0")

    @property
    def waveform(self) -> str:
        return self._waveform

    @waveform.setter
    def waveform(self, value: str):
        if self._enabled_controls is False:
            raise RuntimeError("Controls of WaveGenerator are blocked during loop")
        if value.lower() not in self._waveforms.keys():
            raise ValueError(f"{value.lower()} is not a available waveform.\n"
                             f"Posible values are: {sorted(self._waveforms.keys())}")
        self._waveform = value.lower()

    def get_signal(self) -> np.ndarray:
        fun = partial(
            self._waveforms[self._waveform],
            freq=self._freq,
            phase=self._phase
        )
        signal = make_signal(fun, duration=self._duration, fs=self._fs)
        signal = signal * self._amplitude * (2**15 - 1)
        return signal


class WaveGenerator:
    def __init__(self, fs: int = 44100, duration: int = 60):
        self._duration = duration
        self._fs = fs
        self._channel1 = WaveGeneratorChannel(fs=fs, duration=duration)
        self._channel2 = WaveGeneratorChannel(fs=fs, duration=duration)
        self._wave_loop = None

    @property
    def channel1(self) -> WaveGeneratorChannel:
        return self._channel1

    @property
    def channel2(self) -> WaveGeneratorChannel:
        return self._channel2

    def get_channels_signal(self) -> np.ndarray:
        signal = np.array([
            channel.get_signal()
            for channel in [self._channel1, self._channel2]
            if channel.enabled is True
        ])
        return signal

    def play_loop(self) -> WaveLoop:
        self._wave_loop = WaveLoop(
            signal=self.get_channels_signal(),
            normalize=False,
            fs=self._fs
        )
        return self._wave_loop

    def play_once(self):
        self._wave_loop = WaveLoop(
            signal=self.get_channels_signal(),
            normalize=False,
            fs=self._fs
        )
        self._wave_loop.play_once()


class WaveLoop:
    def __init__(self, signal: np.ndarray, normalize: bool = False, fs: int = 44100):
        self._fs = fs
        self._play = multiprocessing.Event()
        self._force_stop = multiprocessing.Event()
        self._repeat = multiprocessing.Event()
        self._play.clear()
        self._wave_object = self._generate(signal, normalize)
        self._play_object = None
        self._process = multiprocessing.Process(target=self._loop, daemon=True)

    def is_playing(self) -> bool:
        return self._play_object.is_playing()

    def _generate(self, signal: np.ndarray, normalize: bool = True) -> sa.WaveObject:

        signal = np.squeeze(signal)
        if signal.ndim == 1:
            channels = [signal]
        elif signal.ndim == 2:
            if signal.shape[0] == 2:
                channels = [signal[0, :], signal[1, :]]
            else:
                raise ValueError(f"Unexpected shape of signal: "
                                 f"signal.shape={signal.shape}.\n"
                                 f"Posible values are: \n"
                                 f"  (n)    -> single channel\n"
                                 f"  (1, n) -> single channel\n"
                                 f"  (2, n) -> double channel\n")
        else:
            raise ValueError(f"Unexpected dimension of signal: "
                             f"signal.ndim={signal.ndim}.\n"
                             f"Posible values are: 1 or 2")

        if normalize is True:
            # Ensure that highest value is in 16-bit range
            channels = [
                ch * (2**15 - 1) / np.max(np.abs(ch))
                for ch in channels
            ]
        else:
            if np.any(signal < -(2**15 - 1)) or np.any(signal > (2**15 - 1)):
                raise ValueError(f"Signal type is a 16-bit signed integer. "
                                 f"Values under -32767 or upper 32767 are not allowed")

        audio = np.ascontiguousarray(np.asarray(channels).T)
        audio = audio.astype(np.int16)

        wave_object = sa.WaveObject(
            audio_data=audio,
            num_channels=len(channels),
            bytes_per_sample=2,
            sample_rate=self._fs
        )

        return wave_object

    def _loop(self):
        while self._play.is_set():
            self._play_object = self._wave_object.play()
            while self._play_object.is_playing() and not self._force_stop.is_set():
                pass

    def play_once(self):
        try:
            self._play_object = self._wave_object.play()
            self._play_object.wait_done()
        finally:
            self._play_object.stop()

    def __enter__(self) -> WaveLoop:
        self._play.set()
        self._force_stop.clear()
        self._process.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._play.clear()
        self._force_stop.set()
        self._process.join()



from arduscope.arduscope import Arduscope, ArduscopeMeasure
from arduscope.fit_tools import sine, fit_signal
try:
    from arduscope.wave_generator import WaveGenerator, WaveLoop
except ModuleNotFoundError:
    print("  NOTE: WaveGenerator requieres the extra package: simpleaudio\n"
          "      Use `pip install simpleaudio`."
          "      For Linux see simpleaudio dependecies at:\n"
          "      https://simpleaudio.readthedocs.io"
          "/en/latest/installation.html#linux-dependencies")



import numpy as np
from typing import Callable, Tuple
from scipy.fft import rfft, rfftfreq
from scipy.optimize import curve_fit
from scipy.signal import find_peaks
from scipy.signal.signaltools import correlate, correlation_lags
from scipy.interpolate import UnivariateSpline


SignalFunction = Callable[[np.ndarray, float, float, float, float], np.ndarray]


class UnfittableDataError(Exception):
    """The fit can't be performed over the input data."""

    pass


def average_vpp(y: np.ndarray) -> Tuple[float, float]:
    mean_max_vpp = y[y > np.percentile(y, 95)].mean()
    mean_min_vpp = y[y < np.percentile(y, 5)].mean()
    std_max_vpp = y[y > np.percentile(y, 95)].std()
    std_min_vpp = y[y < np.percentile(y, 5)].std()
    mean_vpp = mean_max_vpp - mean_min_vpp
    std_vpp = std_max_vpp + std_min_vpp

    return mean_vpp, std_vpp


def estimate_fft(x: np.ndarray, y: np.ndarray) -> Tuple[float, float]:
    yf = rfft(y)
    xf = rfftfreq(y.size, float(np.mean(x[1:] - x[0:-1])))
    peaks, _ = find_peaks(np.abs(yf), height=np.mean(np.abs(yf)))

    if len(peaks) > 0:
        freq = xf[peaks[0]]
        angle = np.angle(yf[peaks[0]])
        return freq, angle
    else:
        return np.nan, np.nan


def estimate_roots(x: np.ndarray, y: np.ndarray) -> Tuple[float, float]:
    spl = UnivariateSpline(x, y - np.mean(y), s=1 / len(y))
    zeros = spl.roots()

    if len(zeros) > 1:
        freq = float(1 / np.mean(zeros[1:] - zeros[0:-1]) / 2)
        angle = 2 * np.pi * zeros[0] * freq
        angle = (angle - np.pi) % (2 * np.pi) - np.pi
        return freq, angle
    else:
        return np.nan, np.nan


def estimate_correlation(x: np.ndarray, y: np.ndarray) -> Tuple[float, float]:
    xc = correlation_lags(y.size, y.size) * np.mean(x[1:] - x[0:-1])
    yc = correlate(y, y)
    peaks, _ = find_peaks(yc)
    if len(peaks) > 1:
        freq = float(np.mean(1 / (xc[peaks[1:]] - xc[peaks[0:-1]])))
        x_shift = float(xc[np.argmax(yc)])
        angle = 2 * np.pi * x_shift * freq
        angle = (angle - np.pi) % (2 * np.pi) - np.pi
        return freq, angle
    else:
        return np.nan, np.nan


def estimate_phase_correlation(x: np.ndarray, y: np.ndarray, freq: float) -> float:
    mean_vpp, std_vpp = average_vpp(y)
    y_without_phase = sine(x, mean_vpp / 2, freq, 0.0, float(y.mean()))
    xc = correlation_lags(x.size, x.size) * (x[1] - x[0])
    yc = correlate(y, y_without_phase)
    x_shift = xc[np.argmax(yc)]
    angle = 2 * np.pi * x_shift * freq
    return angle


def sine(
    x: np.ndarray, a: float, freq: float, phase: float, offset: float
) -> np.ndarray:
    return a * np.sin(x * 2 * np.pi * freq + phase) + offset


def fit_signal(
    f: SignalFunction, x: np.ndarray, y: np.ndarray
) -> Tuple[np.ndarray, np.ndarray, float]:
    mean_vpp, std_vpp = average_vpp(y)
    estimated_params = np.asarray(
        [estimate_fft(x, y), estimate_roots(x, y), estimate_correlation(x, y)]
    )

    _params, _std, _rmse = [], [], []
    for freq, angle in estimated_params:
        if freq is np.nan or angle is np.nan:
            _params.append(np.nan)
            _std.append(np.nan)
            _rmse.append(np.nan)
        else:
            p0 = [mean_vpp / 2, freq, angle, y.mean()]
            p_opt, p_cov = curve_fit(f, x, y, p0=p0)
            std = np.sqrt(np.diag(p_cov))
            rmse = np.sqrt(np.mean((f(x, *p_opt) - y) ** 2))
            _params.append(p_opt)
            _std.append(std)
            _rmse.append(rmse)

    if np.all(np.isnan(_rmse)):
        raise UnfittableDataError("This data can't be fitted by this method")

    min_index = np.nanargmin(_rmse)
    if _params[min_index][0] < 0:
        _params[min_index][0] = np.abs(_params[min_index][0])
        _params[min_index][2] += np.pi

    _params[min_index][2] = (_params[min_index][2] - np.pi) % (2 * np.pi) - np.pi

    return _params[min_index], _std[min_index], _rmse[min_index]
